from typing import List
import subprocess

from libdskmgr.common import Location
from libdskmgr.wm.wm import WindowManager

class Bspwm(WindowManager):
    @staticmethod
    def _get_desktop_name(location: Location) -> str:
        return f'{location.x}-{location.y}'

    def create_desktop(self, location: Location) -> None:
        subprocess.run(['bspc', 'monitor', '--add-desktops', Bspwm._get_desktop_name(location)])

    def focus_desktop(self, location: Location) -> None:
        subprocess.run(['bspc', 'desktop', '--focus', Bspwm._get_desktop_name(location)])

    def initialize_desktops(self, locations: List[Location]) -> None:
        subprocess.run(['bspc', 'monitor', '--reset-desktops', *map(Bspwm._get_desktop_name, locations)])



import os
from typing import Callable
from socket import socket, AF_UNIX

class UnixSocketServer:
    def __init__(self, socket_path: str, connection_handler: Callable[[socket], None]) -> None:
        self._socket = socket(AF_UNIX)
        self._socket_path = socket_path
        self._connection_handler = connection_handler
        
    def __enter__(self) -> 'UnixSocketServer':
        self._socket.bind(self._socket_path)
        self._socket.listen()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        self._socket.close()
        try:
            os.unlink(self._socket_path)
        except OSError:
            if os.path.exists(self._socket_path):
                print('Error: could not unlink socket file')

    def run(self) -> None:
        while True:
            connection, _ = self._socket.accept()
            with connection:
                self._connection_handler(connection)



from typing import List

from libdskmgr.common import Location

class WindowManager:
    def create_desktop(self, location: Location) -> None:
        raise NotImplementedError

    def focus_desktop(self, location: Location) -> None:
        raise NotImplementedError

    def initialize_desktops(self, locations: List[Location]) -> None:
        raise NotImplementedError



from typing import List

from libdskmgr.common import Location
from libdskmgr.wm.wm import WindowManager

class BspwmMock(WindowManager):
    @staticmethod
    def _get_desktop_name(location: Location) -> str:
        return f'{location.x}-{location.y}'

    def create_desktop(self, location: Location) -> None:
        print(f"Creating {BspwmMock._get_desktop_name(location)}")

    def focus_desktop(self, location: Location) -> None:
        print(f"Focusing {BspwmMock._get_desktop_name(location)}")

    def initialize_desktops(self, locations: List[Location]) -> None:
        print(f"Initializing desktops [{', '.join(map(BspwmMock._get_desktop_name, locations))}]")




from libdskmgr.wm.wm import WindowManager
from libdskmgr.wm.bspwm import Bspwm
from libdskmgr.wm.bspwm_mock import BspwmMock

class WindowManagerNotSupportedError(Exception):
    pass

class WindowManagerFactory:
    def create(self, wm_name: str) -> WindowManager:
        wm = {
            'bspwm': Bspwm,
            'bspwm-mock': BspwmMock,
        }.get(wm_name)
        if wm is None:
            raise WindowManagerNotSupportedError(f'Unsupported window manager {repr(wm_name)}')
        return wm()



import json
from typing import NamedTuple

from libdskmgr.wm.wm import WindowManager
from libdskmgr.common import Location, Direction

class DesktopGroup:
    def __init__(self, size: int = 1, current: int = 0) -> None:
        self.size = size
        self.current = current

class DesktopManager:
    def __init__(self, wm: WindowManager) -> None:
        self._groups = [DesktopGroup()]
        self._focused_group = 0
        self._wm = wm
    
    def _is_in_bounds(self, loc: Location) -> bool:
        return 0 <= loc.x < len(self._groups) and 0 <= loc.y < self._groups[loc.x].size
    
    def reset_desktops(self, num_groups: int) -> None:
        locations = [Location(i, 0) for i in range(num_groups)]
        self._groups = [DesktopGroup() for i in range(num_groups)]
        self.focus_desktop(Location(0, 0))
        self._wm.initialize_desktops(locations)

    def get_focused_group(self) -> int:
        return self._focused_group

    def create_vertical(self, group: int) -> Location:
        if group < 0 or len(self._groups) <= group: return Location(0, 0) # TODO - error, out of bounds!
        new_desktop_location = Location(group, self._groups[group].size)
        self._wm.create_desktop(new_desktop_location)
        self._groups[group].size += 1
        return new_desktop_location

    def create_horizontal(self) -> Location:
        new_desktop_location = Location(len(self._groups), 0)
        self._wm.create_desktop(new_desktop_location)
        self._groups += [DesktopGroup()]
        return new_desktop_location
    
    def focus_desktop(self, location: Location) -> None:
        if not self._is_in_bounds(location): return # TODO - error, out of bounds!
        self._groups[location.x].current = location.y
        self._focused_group = location.x
        self._wm.focus_desktop(location)
    
    def focus_group(self, group: int) -> None:
        if group < 0 or len(self._groups) <= group: return # TODO - error, out of bounds!
        self.focus_desktop(Location(group, self._groups[group].current))

    def dump_state(self) -> str:
        return json.dumps({
            'focused': self._focused_group,
            'groups': [{'size': group.size, 'y': group.current} for group in self._groups]
        }) + '\n'

    def move(self, direction: Direction) -> None:
        if direction in [Direction.UP, Direction.DOWN]: # move vertically
            vertical_offset = 1 if direction == Direction.UP else -1
            focused_group = self._groups[self._focused_group]
            new_y = (focused_group.current + vertical_offset) % focused_group.size
            self.focus_desktop(Location(self._focused_group, new_y))
        else: # move horizontally
            horizontal_offset = 1 if direction == Direction.RIGHT else -1
            new_x = (self._focused_group + horizontal_offset) % len(self._groups)
            self.focus_group(new_x)



from libdskmgr.server import UnixSocketServer
from libdskmgr.desktop_manager import DesktopManager
from libdskmgr.wm.wm_factory import WindowManagerFactory
from libdskmgr.connections_handler import DskmgrConnectionsHandler

def main() -> int:
    desktop_manager = DesktopManager(WindowManagerFactory().create('bspwm-mock'))
    connections_handler = DskmgrConnectionsHandler(desktop_manager)
    with UnixSocketServer('/tmp/dskmgr_socket', connections_handler.handle_connection) as server:
        server.run()
    return 1

if __name__ == '__main__':
    main()



from socket import socket
from typing import List, Callable

from libdskmgr.common import Direction
from libdskmgr.desktop_manager import DesktopManager

class DskmgrConnectionsHandler:
    def __init__(self, desktop_manager: DesktopManager, max_recv_size = 4096) -> None:
        self._desktop_manager = desktop_manager
        self._max_recv_size = max_recv_size

    def _handle_new_x(self, args: List[str], connection: socket) -> None:
        self._desktop_manager.create_horizontal()
    
    def _handle_new_y(self, args: List[str], connection: socket) -> None:
        if len(args) == 1:
            group = self._desktop_manager.get_focused_group()
        elif len(args) == 2 and args[1].isnumeric():
            group = int(args[1])
        else:
            connection.sendall('Invalid arguments\n'.encode())
            return
        self._desktop_manager.create_vertical(group)

    def _handle_dump(self, args: List[str], connection: socket) -> None:
        connection.sendall(self._desktop_manager.dump_state().encode())

    def _handle_help(self, args: List[str], connection: socket) -> None:
        connection.sendall('TODO - usage\n'.encode())

    def _handle_move(self, args: List[str], connection: socket) -> None:
        DIRECTIONS = {
            'up':    Direction.UP,
            'down':  Direction.DOWN,
            'left':  Direction.LEFT,
            'right': Direction.RIGHT,
        }
        if len(args) != 2 or args[1] not in DIRECTIONS:
            connection.sendall('Invalid arguments\n'.encode())
        else:
            self._desktop_manager.move(DIRECTIONS[args[1]])

    def _handle_reset(self, args: List[str], connection: socket) -> None:
        if len(args) == 2 and args[1].isnumeric():
            self._desktop_manager.reset_desktops(int(args[1]))
        else:
            connection.sendall('Invalid arguments\n'.encode())
        
    def _handle_focus_group(self, args: List[str], connection: socket) -> None:
        if len(args) == 2 and args[1].isnumeric():
            self._desktop_manager.focus_group(int(args[1]))
        else:
            connection.sendall('Invalid arguments\n'.encode())

    def _get_handler(self, args: List[str]) -> Callable[[List[str], socket], None]:
        return self._handle_help if len(args) == 0 else {
            'reset': self._handle_reset,
            'new-x': self._handle_new_x,
            'new-y': self._handle_new_y,
            'dump': self._handle_dump,
            'move': self._handle_move,
            'goto': self._handle_focus_group,
        }.get(args[0]) or self._handle_help

    def handle_connection(self, connection: socket) -> None:
        data = connection.recv(self._max_recv_size)
        print('Recieved:', data)
        args = [arg.decode() for arg in data.split(b'\0') if arg]
        print('Args:', args)
        self._get_handler(args)(args, connection)
